# Copyright (c) 2002 Gavin Brown. All rights reserved. This program is
# free software; you can redistribute it and/or modify it under the same
# terms as Perl itself. 
#
# $Id: Crontab.pm,v 1.25 2003/05/25 15:17:11 jodrell Exp $

package Gnome::Crontab;
use vars qw($VERSION $SNAME $AUTHORS);
use Gnome;
use Gnome::Crontab::L10N;
use Gtk::HandyCList;
use strict;

$VERSION	= '0.0.4';
$SNAME		= 'gnome-crontab';
$AUTHORS	=	[
				'Gavin Brown <gavin.brown@uk.com> (main development)',
				'Dario Conigliaro <mad.horse@tiscalinet.it> (Italian Translation)',
				'Cyril Sudre <cyril.sudre@laposte.net> (French Translation)',
				'Santiago Martínez <smvarela@eresmas.net> (Spanish and Catalan Translation)',
				'Christian Neumair <chris@gnome-de.org> (German Translation)',
				'Hu ChangSong <gugong@21cn.com> (Chinese Traditional and Simplified Translations)',
				'Ysbrand Galama <ysbrand@galama.net> (Dutch Translation)',
			];
sub new {
	my $package = shift;
	my $gc = {};
	bless($gc, $package);
	return $gc;
}

sub init {
	my ($gc, $filename) = @_;

	Gnome->init($SNAME);

	$gc->{lexicon} = Gnome::Crontab::L10N->get_handle();

	$gc->{filename} = $filename;

	$gc->build_ui();
	$gc->build_menu();
	$gc->build_toolbar();

	$gc->reload();

	$gc->{app}->show_all() if ($gc->{filename} ne '');

	Gtk->main();
	return;
}

sub build_ui {
	my $gc = shift;

	$gc->{app} = Gnome::App->new($SNAME, $SNAME);
	my $suffix;
	if ($gc->{url} ne '') {
		$suffix = ": $gc->{url}";
	} else {
		chomp(my $hostname = `hostname`);
		$suffix = ': '.(getpwuid($>))[0].'@'.$hostname;
	}
	$gc->{app}->set_title($gc->string('window_title').$suffix);
	$gc->{app}->set_default_size(300, 500);
	$gc->{app}->signal_connect('delete_event', sub { $gc->discard_close() });

	$gc->{list} = Gtk::HandyCList->new(
				$gc->string('mins'),
				$gc->string('hours'),
				$gc->string('days'),
				$gc->string('months'),
				$gc->string('weekdays'),
				$gc->string('job'),
			);
	$gc->{list}->signal_connect('select_row', sub { $gc->select_row($_[1], $_[3]) });
	$gc->{list}->signal_connect('button_press_event', sub { $gc->button_press(@_) });

	$gc->{scrolled_window} = Gtk::ScrolledWindow->new(undef, undef);
	$gc->{scrolled_window}->set_policy('automatic', 'automatic');
	$gc->{scrolled_window}->add($gc->{list});

	# yes, I know a table would be better here. But I couldn't get the table I built to be non-homogeneous:
	$gc->{env}{data_vbox} = Gtk::VBox->new(0, 0);
	$gc->{env}{data_vbox}->set_spacing(8);
	$gc->{env}{titles_hbox} = Gtk::HBox->new(0, 0);
	$gc->{env}{titles_hbox}->pack_start($gc->label('env_name_title'), 1, 1, 0);
	$gc->{env}{titles_hbox}->pack_start($gc->label('env_value_title'), 1, 1, 0);
	$gc->{env}{data_vbox}->pack_start($gc->{env}{titles_hbox}, 0, 0, 0);

	$gc->{env}{add_button} = $gc->stock_button('Add', $gc->string('add_env_entry'));
	$gc->{env}{add_button}->signal_connect('clicked', sub { $gc->add_blank_env_row() });

	$gc->{env}{vbox} = Gtk::VBox->new(0, 0);
	$gc->{env}{vbox}->set_spacing(8);
	$gc->{env}{vbox}->border_width(8);

	$gc->{env}{scrolled} = Gtk::ScrolledWindow->new();
	$gc->{env}{scrolled}->set_policy('never', 'automatic');
	$gc->{env}{scrolled}->add_with_viewport($gc->{env}{data_vbox});
	$gc->{env}{scrolled}->set_placement('top_left');
	$gc->{env}{scrolled}->child()->set_shadow_type('none');

	$gc->{env}{vbox}->pack_start($gc->{env}{scrolled}, 1, 1, 0);
	$gc->{env}{vbox}->pack_end($gc->{env}{add_button}, 0, 0, 0);

	$gc->{notebook} = Gtk::Notebook->new();
	$gc->{notebook}->append_page($gc->{scrolled_window}, $gc->label('schedule'));
	$gc->{notebook}->append_page($gc->{env}{vbox}, $gc->label('environment'));

	$gc->{mailto} = Gtk::Entry->new();
	$gc->{hbox} = Gtk::HBox->new(0, 0);
	$gc->{hbox}->pack_start($gc->label('report_to'), 0, 0, 0);
	$gc->{hbox}->pack_start($gc->{mailto}, 1, 1, 0);

	$gc->{action_area} = Gtk::HButtonBox->new();
	$gc->{action_area}->set_layout_default('edge');

	$gc->{buttons}{reload_button} = $gc->stock_button('Refresh', $gc->string('button_reload'));
	$gc->{buttons}{reload_button}->signal_connect('clicked', sub { $gc->reload() });

	$gc->{buttons}{ok_button} = $gc->stock_button('Button_Ok', $gc->string('button_save'));
	$gc->{buttons}{ok_button}->signal_connect('clicked', sub { $gc->save_close() ; exit });

	$gc->{buttons}{cancel_button} = $gc->stock_button('Button_Cancel', $gc->string('button_discard'));
	$gc->{buttons}{cancel_button}->signal_connect('clicked', sub { $gc->discard_close() });

	$gc->{action_area}->add($gc->{buttons}{reload_button});
	$gc->{action_area}->add($gc->{buttons}{ok_button});
	$gc->{action_area}->add($gc->{buttons}{cancel_button});

	$gc->{vbox} = Gtk::VBox->new(0, 0);
	$gc->{vbox}->set_border_width(8);
	$gc->{vbox}->set_spacing(8);
	$gc->{vbox}->pack_start($gc->{notebook}, 1, 1, 0);
	$gc->{vbox}->pack_start($gc->{hbox}, 0, 1, 0);
	$gc->{vbox}->pack_start(Gtk::HSeparator->new(), 0, 1, 0);
	$gc->{vbox}->pack_start($gc->{action_area}, 0, 1, 0);

	$gc->{app}->set_contents($gc->{vbox});

	return;
}

sub build_menu {
	my $gc = shift;
	$gc->{app}->create_menus(
		{
			type	=> 'subtree',
			label	=> $gc->string('file'),
			subtree	=> 	[
						{
							type		=> 'item',
							label		=> $gc->string('remote_connection'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Exec',
							callback	=> sub { system("$0 remote &")}
						},
						{	type		=> 'separator' },
						{
							type		=> 'item',
							label		=> $gc->string('reload'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Refresh',
							callback	=> sub { $gc->reload() }
						},
						{
							type		=> 'item',
							label		=> $gc->string('save'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Button_Ok',
							callback	=> sub { $gc->save_close() ; exit }
						},
						{
							type		=> 'item',
							label		=> $gc->string('discard'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Button_Cancel',
							callback	=> sub { $gc->discard_close() }
						}
					]
		},
		{
			type	=> 'subtree',
			label	=> $gc->string('edit'),
			subtree	=>	[
						{
							type		=> 'item',
							label		=> $gc->string('new_entry'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Menu_New',
							callback	=> sub { $gc->new_entry() }
						},
						{
							type		=> 'item',
							label		=> $gc->string('edit_entry'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Menu_Properties',
							callback	=> sub { $gc->edit_entry() }
						},
						{
							type		=> 'item',
							label		=> $gc->string('delete_entry'),
							pixmap_type	=> 'stock',
							pixmap_info	=> 'Menu_Trash',
							callback	=> sub { $gc->delete_entry() }
						},
					]
							
		},
		{
			type	=> 'subtree',
			label	=> $gc->string('menu_help'),
			subtree	=>	[
						{
							type		=> 'item',
							label		=> $gc->string('help'),
							pixmap_type 	=> 'stock',
							pixmap_info	=> 'Menu_Help',
							callback	=> sub { $gc->help() }
						},
						{	type		=> 'separator' },
						{
							type		=> 'item',
							label		=> $gc->string('about'),
							pixmap_type 	=> 'stock',
							pixmap_info	=> 'Menu_About',
							callback	=> sub { $gc->about() }
						}
					]
		}
	);
	return;
}

sub build_toolbar {
	my $gc = shift;
	$gc->{app}->create_toolbar(
		{
			type		=> 'item',
			label		=> $gc->string('new_entry_button'),
			pixmap_type	=> 'stock',
			pixmap_info	=> 'New',
			callback	=> sub { $gc->new_entry() },
			hint		=> $gc->string('button_new_hint'),
		},
		{
			type		=> 'item',
			label		=> $gc->string('edit_entry_button'),
			pixmap_type	=> 'stock',
			pixmap_info	=> 'Properties',
			callback	=> sub { $gc->edit_entry() },
			hint		=> $gc->string('button_edit_hint'),
		},
		{
			type		=> 'item',
			label		=> $gc->string('delete_entry_button'),
			pixmap_type	=> 'stock',
			pixmap_info	=> 'Trash',
			callback	=> sub { $gc->delete_entry() },
			hint		=> $gc->string('button_delete_hint'),
		},
	);
	return;
}

sub string {
	my ($gc, $string) = @_;
	return $gc->{lexicon}->maketext($string);
}

sub stock_button {
	my ($gc, $icon, $string) = @_;
	return Gnome::Stock->gnome_pixmap_button(Gnome::Stock->new_with_icon($icon), $string);
}

sub parse_crontab {
	my $gc = shift;
	use Data::Dumper;
	$gc->{items} = [];
	open(FILE, $gc->{filename}) or $gc->error("$gc->{filename}: $!") and return undef;
	while (<FILE>) {
		chomp;
		s/\s+/ /g;
		unless ($_ eq '' || $_ =~ /^\n/ || $_ =~ /^\s*#/) {
			if (/^MAILTO=\"*(.+?)\"*$/i) {
				$gc->mailto($1);
			} elsif (/([a-z0-9_]+)=\"*(.+?)\"*$/i) {
				$gc->{environment}{$1} = $2;
			} else {
				my @entry = split(/\s+/, $_, 6);
				$gc->add_item(\@entry);
			}
		}
	}
	close(FILE);
	return;
}

sub add_item {
	my ($gc, $item) = @_;
	push(@{$gc->{items}}, $item);
	return;
}

sub update_list {
	my $gc = shift;
	$gc->{list}->clear();
	@{$gc->{items}} = grep { defined } @{$gc->{items}};
	$gc->{list}->data(@{$gc->{items}});
	return;
}

sub error {
	my ($gc, $str) = @_;
	$gc->{app}->error($str);
	return;
}

sub message {
	my ($gc, $str) = @_;
	$gc->{app}->message($gc->string($str));
	return;
}

sub reload {
	my $gc = shift;
	$gc->parse_crontab();
	$gc->update_list();
	$gc->set_environment();
}

sub save_close {
	my $gc = shift;
	open(FILE, ">$gc->{filename}") or $gc->error("$gc->{filename}: $!") and return undef;
	printf(FILE "\n\n### crontab file generated by %s on %s.\n\n", $SNAME, scalar(localtime()));

	printf(FILE "MAILTO=\"%s\"\n\n", $gc->mailto()) if ($gc->mailto() ne '');

	if (defined($gc->{env}{name_entries})) {
		print FILE "### Begin environment section:\n";
		for (my $i = 0 ; $i < scalar(@{$gc->{env}{name_entries}}) ; $i++) {
			if (defined($gc->{env}{name_entries}[$i])) {
				my $name  = $gc->{env}{name_entries}[$i]->get_text();
				my $value = $gc->{env}{value_entries}[$i]->get_text();
				print FILE "$name=$value\n" if ($name ne '' && $value ne '');
			}
		}
		print FILE "### End environment section\n\n";
	}

	print FILE "### Begin Schedule section:\n";
	foreach my $entry (@{$gc->{items}}) {
		printf(	FILE
			"%s\t%s\t%s\t%s\t%s\t%s\n",
			$entry->{$gc->string('mins')},
			$entry->{$gc->string('hours')},
			$entry->{$gc->string('days')},
			$entry->{$gc->string('months')},
			$entry->{$gc->string('weekdays')},
			$entry->{$gc->string('job')},
		);
	}
	close(FILE);
	return;
}

sub discard_close {
	exit;
}

sub edit_entry {
	my $gc = shift;
	my $entry_id = $gc->current_selection();
	$gc->edit_window($entry_id);
	return;
}

sub delete_entry {
	my $gc = shift;
	my $entry_id = $gc->current_selection();
	return if ($entry_id eq '');
	@{$gc->{items}}[$entry_id] = undef;
	$gc->{list}->{handy}{selection} = {};
	$gc->update_list();
	return;
}

sub new_entry {
	my $gc = shift;
	$gc->edit_window();
	return;
}

sub help {
	my $gc = shift;
	$gc->error($gc->string('help_coming_soon'));
	return;
}

sub about {
	my $gc = shift;
	$gc->{about} = Gnome::About->new(
		$gc->string('window_title'),
		$VERSION,
		$gc->string('comments'),
		$AUTHORS,
		$gc->string('copyright'),
	);
	$gc->{about}->show();
	return;
}

sub edit_window {
	my ($gc, $entry_id) = @_;
	my $entry = (defined($entry_id) ? @{$gc->{items}}[$entry_id] : {});
	my $gce = {};
	$gce->{window} = Gtk::Window->new('toplevel');
	$gce->{window}->signal_connect('delete_event', sub { $gce->{window}->destroy() });
	$gce->{window}->set_title($gc->string('edit_window_title'));

	$gce->{command}{frame} = Gtk::Frame->new($gc->string('command_frame'));

	$gce->{command}{input} = Gtk::Entry->new();
	$gce->{command}{input}->set_text($entry->{$gc->string('job')});

	$gce->{command}{browse} = $gc->stock_button('Menu_Search', $gc->string('command_browse'));
	$gce->{command}{browse}->signal_connect('clicked', sub { $gc->command_browse($gce) });

	$gce->{command}{hbox} = Gtk::HBox->new(0, 0);
	$gce->{command}{hbox}->set_spacing(8);
	$gce->{command}{hbox}->set_border_width(8);
	$gce->{command}{hbox}->pack_start($gce->{command}{input},  1, 1, 0);
	$gce->{command}{hbox}->pack_start($gce->{command}{browse}, 0, 0, 0);

	$gce->{command}{frame}->add($gce->{command}{hbox});

	$gce->{schedule}{frame} = Gtk::Frame->new($gc->string('schedule_frame'));
	$gce->{schedule}{table} = Gtk::Table->new(5, 3);
	$gce->{schedule}{table}->set_border_width(8);
	$gce->{schedule}{table}->set_row_spacings(8);
	$gce->{schedule}{table}->set_col_spacings(8);

	my $row = 0;
	foreach my $key (qw(mins hours days months weekdays)) {
		my $value = $entry->{$gc->string($key)};
		$gce->{schedule}{$key}{label} = $gc->label($key);
		$gce->{schedule}{$key}{input} = Gtk::Entry->new();
		$gce->{schedule}{$key}{input}->set_text($value);
		$gce->{schedule}{$key}{button} = $gc->stock_button('Menu_Properties', $gc->string('wizard'));
		$gce->{schedule}{$key}{button}->signal_connect('clicked', sub { $gc->wizard($gce, $key, $gce->{schedule}{$key}{input}->get_text()) });
		$gce->{schedule}{table}->attach_defaults($gce->{schedule}{$key}{label}, 0, 1, $row, $row+1);
		$gce->{schedule}{table}->attach_defaults($gce->{schedule}{$key}{input}, 1, 2, $row, $row+1);
		$gce->{schedule}{table}->attach_defaults($gce->{schedule}{$key}{button}, 2, 3, $row, $row+1);
		$row++;
	}

	$gce->{schedule}{frame}->add($gce->{schedule}{table});

	$gce->{buttons}{ok_button} = $gc->stock_button('Button_Ok', $gc->string('ok'));
	$gce->{buttons}{ok_button}->signal_connect('clicked',	sub {
									my $entry =	[
												$gce->{schedule}{mins}{input}->get_text(),
												$gce->{schedule}{hours}{input}->get_text(),
												$gce->{schedule}{days}{input}->get_text(),
												$gce->{schedule}{months}{input}->get_text(),
												$gce->{schedule}{weekdays}{input}->get_text(),
												$gce->{command}{input}->get_text(),
											];
									if (defined($entry_id)) {
										@{$gc->{items}}[$entry_id] = $entry;
									} else {
										push(@{$gc->{items}}, $entry);
									}
									$gc->update_list();
									$gce->{window}->destroy();
								});
	$gce->{buttons}{cancel_button} = $gc->stock_button('Button_Cancel', $gc->string('cancel'));
	$gce->{buttons}{cancel_button}->signal_connect('clicked', sub { $gce->{window}->destroy() });

	$gce->{action_area} = Gtk::HButtonBox->new();
	$gce->{action_area}->set_layout_default('edge');
	$gce->{action_area}->add($gce->{buttons}{ok_button});
	$gce->{action_area}->add($gce->{buttons}{cancel_button});

	$gce->{vbox} = Gtk::VBox->new(0, 0);
	$gce->{vbox}->set_border_width(8);
	$gce->{vbox}->set_spacing(8);

	$gce->{vbox}->pack_start($gce->{command}{frame},  0, 0, 0);
	$gce->{vbox}->pack_start($gce->{schedule}{frame}, 1, 1, 0);
	$gce->{vbox}->pack_start(Gtk::HSeparator->new(),  0, 0, 0);
	$gce->{vbox}->pack_start($gce->{action_area},     0, 0, 0);

	$gce->{window}->add($gce->{vbox});

	$gce->{window}->show_all();
	return;
}

sub current_selection {
	my $gc = shift;
	my @selected = keys %{$gc->{list}->selection()};
	my $entry_id = $selected[scalar(@selected) - 1];
	return $entry_id;
}

sub label {
	my ($gc, $str) = @_;
	my $label = Gtk::Label->new($gc->string($str));
	$label->set_justify('left'),
	$label->set_line_wrap(1);
	return $label;
}

sub mailto {
	my ($gc, $str) = @_;
	if ($str ne '') {
		$gc->{mailto}->set_text($str);
		return;
	} else {
		return $gc->{mailto}->get_text();
	}
}

sub select_row {
	my ($gc, $entry_id, $event) = @_;
	if ($event->{type} eq '2button_press') {
		$gc->edit_window($entry_id);
	}
	return;
}

sub button_press {
	my ($gc, $list, $event) = @_;
	if ($event->{button} == 3) {
		$gc->popup($event->{button});
	}
	return;
}

sub popup {
	my ($gc, $button) = @_;

	return if ($gc->current_selection() eq '');

	my $menu = Gtk::Menu->new();
	$menu->set_title($gc->string('popup_title'));
	$menu->popup(	undef,
			undef,
			0,
			$button,
			undef);

	my $new_item = Gtk::MenuItem->new($gc->string('button_new_hint'));
	$new_item->signal_connect('activate', sub { $gc->new_entry() });
	$menu->append($new_item);

	my $edit_item = Gtk::MenuItem->new($gc->string('button_edit_hint'));
	$edit_item->signal_connect('activate', sub { $gc->edit_entry() });
	$menu->append($edit_item);

	my $delete_item = Gtk::MenuItem->new($gc->string('button_delete_hint'));
	$delete_item->signal_connect('activate', sub { $gc->delete_entry() });
	$menu->append($delete_item);

	$menu->show_all();

	return;
}

sub command_browse {
	my ($gc, $gce) = @_;
	my $selector = Gtk::FileSelection->new($gc->string('browse_title'));
	$selector->ok_button->signal_connect('clicked',	sub {
								my $filename = $selector->get_filename();
								$gce->{command}{input}->set_text($filename);
								$selector->destroy();
								$gc->message('not_executable') unless (-x $filename);
							});
	$selector->cancel_button->signal_connect('clicked', sub { $selector->destroy() });
	$selector->show_all();
	return;
}

sub wizard {
	my ($gc, $gce, $field, $value) = @_;
	my $gcw = {};
	$gcw->{window} = Gtk::Window->new('toplevel');
	$gcw->{window}->signal_connect('delete_event', sub { $gcw->{window}->destroy() });
	$gcw->{window}->set_title($gc->string('wizard_title').$gc->string($field));

	use Gnome::Crontab::Minutes;
	use Gnome::Crontab::Hours;
	use Gnome::Crontab::Days;
	use Gnome::Crontab::Months;
	use Gnome::Crontab::WeekDays;

	SWITCH: {
		$field eq 'mins'	and $gcw->{wizard} = Gnome::Crontab::Minutes->new($gc,  $value),last SWITCH;
		$field eq 'hours'	and $gcw->{wizard} = Gnome::Crontab::Hours->new($gc,    $value),last SWITCH;
		$field eq 'days'	and $gcw->{wizard} = Gnome::Crontab::Days->new($gc,     $value),last SWITCH;
		$field eq 'months'	and $gcw->{wizard} = Gnome::Crontab::Months->new($gc,   $value),last SWITCH;
		$field eq 'weekdays'	and $gcw->{wizard} = Gnome::Crontab::WeekDays->new($gc, $value),last SWITCH;
	}

	$gcw->{buttons}{ok_button} = $gc->stock_button('Button_Ok', $gc->string('ok'));
	$gcw->{buttons}{ok_button}->signal_connect('clicked',	sub {
									$gce->{schedule}{$field}{input}->set_text($gcw->{wizard}->value());
									$gcw->{window}->destroy();
								});
	$gcw->{buttons}{cancel_button} = $gc->stock_button('Button_Cancel', $gc->string('cancel'));
	$gcw->{buttons}{cancel_button}->signal_connect('clicked', sub { $gcw->{window}->destroy() });

	$gcw->{action_area} = Gtk::HButtonBox->new();
	$gcw->{action_area}->set_layout_default('edge');
	$gcw->{action_area}->add($gcw->{buttons}{ok_button});
	$gcw->{action_area}->add($gcw->{buttons}{cancel_button});

	$gcw->{vbox} = Gtk::VBox->new(0, 0);
	$gcw->{vbox}->set_border_width(8);
	$gcw->{vbox}->set_spacing(8);

	$gcw->{vbox}->pack_start($gcw->{wizard}->{widget}, 1, 1, 0);
	$gcw->{vbox}->pack_start(Gtk::HSeparator->new(), 0, 0, 0);
	$gcw->{vbox}->pack_start($gcw->{action_area}, 0, 0, 0);

	$gcw->{window}->add($gcw->{vbox});
	$gcw->{window}->show_all();

	return;
}

sub set_environment {
	my $gc = shift;
	$gc->reset_environment();
	foreach my $name (keys %{$gc->{environment}}) {
		$gc->add_env_row($name, $gc->{environment}{$name});
	}
	return;
}

sub reset_environment {
	my $gc = shift;
	if (defined($gc->{env}{name_entries})) {
		for (my $i = 0 ; $i < scalar(@{$gc->{env}{name_entries}}) ; $i++) {
			${$gc->{env}{name_entries}}[$i]->destroy();
			${$gc->{env}{value_entries}}[$i]->destroy();
			${$gc->{env}{vboxes}}[$i]->destroy();
		}
	}
	return;
}

sub add_blank_env_row {
	my $gc = shift;
	return $gc->add_env_row(undef, undef);
}

sub add_env_row {
	my ($gc, $name, $value) = @_;

	my $name_entry = Gtk::Entry->new();
	$name_entry->set_text($name);

	my $id;
	if (defined($gc->{env}{name_entries})) {
		$id = scalar @{$gc->{env}{name_entries}};
	} else {
		$id = 0;
	}

	push(@{$gc->{env}{name_entries}}, $name_entry);

	my $value_entry = Gtk::Entry->new();
	$value_entry->set_text($value);
	push(@{$gc->{env}{value_entries}}, $value_entry);

	my $delete_button = $gc->stock_button('Remove', $gc->string('remove_env_entry'));
	$delete_button->signal_connect('clicked', sub {
		${$gc->{env}{name_entries}}[$id]->destroy();
		${$gc->{env}{value_entries}}[$id]->destroy();
		${$gc->{env}{vboxes}}[$id]->destroy();
		splice(@{$gc->{env}{name_entries}}, $id, 1, undef);
		splice(@{$gc->{env}{name_entries}}, $id, 1, undef);
		splice(@{$gc->{env}{vboxes}}, $id, 1, undef);
	});

	my $hbox = Gtk::HBox->new(0, 0);
	$hbox->set_spacing(8);
	$hbox->pack_start($name_entry,    1, 1, 0);
	$hbox->pack_start($value_entry,   1, 1, 0);
	$hbox->pack_start($delete_button, 0, 0, 0);
	$hbox->show_all();
	$name_entry->grab_focus();
	push(@{$gc->{env}{vboxes}}, $hbox);

	$gc->{env}{data_vbox}->pack_start($hbox, 0, 0, 0);

	return;
}

# Gnome::App's request_string() is broken when you hit cancel. So here's a better one:
sub request_string {
	my ($gc, $prompt, $ok_callback, $cancel_callback) = @_;
	die "usage: request_string(gc, prompt, ok_callback, cancel_callback)" if ($gc eq '' || $prompt eq '' || $ok_callback eq '' || $cancel_callback eq '');
	my $dialog = Gtk::Dialog->new();
	$dialog->signal_connect('delete_event', $cancel_callback);
	$dialog->set_position('center');
	$dialog->set_modal(1);
	$dialog->set_title($prompt);
	$dialog->vbox()->set_border_width(8);
	$dialog->vbox()->set_spacing(8);
	$dialog->action_area()->set_border_width(8);
	$dialog->action_area()->set_spacing(8);

	$dialog->vbox()->pack_start(Gtk::Label->new($prompt), 1, 1, 0);

	my $entry = Gtk::Entry->new();
	$entry->signal_connect('activate', sub { $dialog->destroy() ; &$ok_callback($entry->get_text()) });

	$dialog->vbox()->pack_start($entry, 1, 1, 0);

	my $ok_button = $gc->stock_button('Button_Ok', $gc->string('ok'));
	$ok_button->signal_connect('clicked', $ok_callback, sub { $dialog->destroy() ; &$ok_callback($entry->get_text()) });

	my $cancel_button = $gc->stock_button('Button_Cancel', $gc->string('cancel'));
	$cancel_button->signal_connect('clicked', sub { $dialog->destroy() ; &$cancel_callback() });

	$dialog->action_area()->pack_start($ok_button, 1, 1, 0);
	$dialog->action_area()->pack_start($cancel_button, 1, 1, 0);

	$dialog->show_all();
	$entry->grab_focus();

	return;
}

1;
