# Copyright (c) 2002 Gavin Brown. All rights reserved. This program is
# free software; you can redistribute it and/or modify it under the same
# terms as Perl itself. 
#
# $Id: Days.pm,v 1.1 2002/11/27 14:57:56 jodrell Exp $

package Gnome::Crontab::Days;

sub new {
	my ($class, $gc, $value) = @_;
	my $gcm = {};
	bless($gcm, $class);

	if ($value =~ /\*/) {
		$gcm->{wildcard} = 1;
	} else {
		foreach my $day (split(/,/, $value)) {
			$day = sprintf('%02d', $day);
			$gcm->{values}{$day} = 1;
		}
	}

	$gcm->{widget} = Gtk::Table->new(6, 7);
	$gcm->{widget}->set_row_spacings(8);
	$gcm->{widget}->set_col_spacings(8);

	my @days = 1..31;
	my $row = 0;
	while (scalar(@days) > 0) {
		my @row = splice(@days, 0, 7);
		my $column = 0;
		foreach my $day (@row) {
			$day = sprintf('%02d', $day);
			$gcm->{checkboxes}{$day} = Gtk::CheckButton->new($day);
			$gcm->{checkboxes}{$day}->set_active(1) if ($gcm->{values}{$day} == 1 || $gcm->{wildcard} == 1);
			$gcm->{checkboxes}{$day}->signal_connect('toggled', sub { $gcm->day_toggle($day) });
			$gcm->{widget}->attach_defaults($gcm->{checkboxes}{$day}, $column, $column+1, $row, $row+1);
			$column++;
		}
		$row++;
	}

	$gcm->{toggle} = Gtk::CheckButton->new($gc->string('select_all'));
	$gcm->{toggle}->set_active(1) if ($gcm->{wildcard} == 1);

	$gcm->{toggle}->signal_connect('toggled', sub { $gcm->select_all_toggle($gc) });

	$gcm->{widget}->attach_defaults($gcm->{toggle}, 0, 7, 5, 6);

	return $gcm;
}

sub day_toggle {
	my ($gcm, $day) = @_;

	if (!$gcm->{checkboxes}{$day}->get_active()) {
		undef $gcm->{wildcard};
		$gcm->{toggle}->set_active(0);
	}

	return;
}

sub select_all_toggle {
	my ($gcm, $gc) = @_;
	if ($gcm->{toggle}->get_active()) {
		for ($i = 1 ; $i <= 31 ; $i++) {
			my $day = sprintf('%02d', $i);
			$gcm->{checkboxes}{$day}->set_active(1);
		}
		$gcm->{wildcard} = 1;
	}
	return;
}

sub value {
	my $gcm = shift;

	if ($gcm->{wildcard} == 1) {
		return '*';
	} else {
		my @return;
		for ($i = 1 ; $i <= 31 ; $i++) {
			my $day = sprintf('%02d', $i);
			push(@return, $day) if ($gcm->{checkboxes}{$day}->get_active());
		}
		return join(',', sort(@return));
	}

}

1;