# $Id: Remote.pm,v 1.3 2003/04/29 15:24:30 jodrell Exp $
package Gnome::Crontab::Remote;
use Gnome::Crontab;
@ISA = qw(Gnome::Crontab);

sub new {
	my $package = shift;
	my $gc = {};
	chomp($gc->{ssh} = `which ssh`);
	bless($gc, $package);
	return $gc;
}

# this overrides the parse_crontab in Crontab.pm. It basically just wraps
# the original function with some ssh-specific stuff:
sub parse_crontab {
	my $gc = shift;
	# we've already asked the user for a URL, so just re-use the tempfile we've got:
	if ($gc->{url} ne '') {
		# use Gnome::Crontab's function to parse the crontab:
		Gnome::Crontab::parse_crontab($gc);
		$gc->update_list();
		$gc->set_environment();
		$gc->{app}->show_all();
	} else {
		$gc->request_string($gc->string('enter_username_host'), sub {
			$gc->{url} = shift;
			$gc->{url} =~ s/\s//g;
			my ($user, $host) = split(/@/, $gc->{url}, 2);
			$user =~ s/[^A-Za-z0-9_\.]//g;
			$host =~ s/[^A-Za-z0-9_\.\-]//g;
			$host = 'localhost' if ($host eq '');
			$gc->{url} = lc("$user\@$host");
			if ($gc->{url} eq '') {
				$gc->error($gc->string('invalid_url'));
				return undef;
			}
			my $cmd = sprintf("%s %s 'crontab -l'", $gc->{ssh}, $gc->{url});
			$gc->{filename} = sprintf('/tmp/gnome-crontab-%s-%d', (getpwuid($>))[0], time());
			open(SSHCMD, "$cmd|") or $gc->error("$cmd: $!") and return undef;
			open(TMPFILE, ">$gc->{filename}") or $gc->error("$gc->{filename}: $!") and return undef;
			while (<SSHCMD>) {
				Gtk->main_iteration() while (Gtk->events_pending());
				print TMPFILE $_;
			}
			close(TMPFILE);
			chmod(0600, $gc->{filename});
			close(SSHCMD);
			Gnome::Crontab::parse_crontab($gc);
			$gc->update_list();
			$gc->set_environment();
			$gc->{app}->set_title($gc->string('window_title').": $gc->{url}");
			$gc->{app}->show_all();
		}, sub {
			$gc->discard_close();
		});
	}
	return;
}

sub reload {
	my $gc = shift;
	$gc->parse_crontab();
}

sub discard_close {
	my $gc = shift;
	unlink($gc->{filename});
	exit;
}

sub save_close {
	my $gc = shift;
	Gnome::Crontab::save_close($gc);
	my $cmd = sprintf("%s %s 'cat >> %s-remote ; crontab %s-remote' ; rm -f %s-remote", $gc->{ssh}, $gc->{url}, $gc->{filename}, $gc->{filename}, $gc->{filename});
	open(SSHCMD, "|$cmd") or $gc->error("$cmd: $!") and return undef;
	open(TMPFILE, $gc->{filename}) or $gc->error("$gc->{filename}: $!") and return undef;
	while (<TMPFILE>) {
		Gtk->main_iteration() while (Gtk->events_pending());
		print SSHCMD $_;
	}
	close(TMPFILE);
	close(SSHCMD);
	unlink($gc->{filename}) or $gc->error("unlink($gc->{filename}): $!") and return undef;;
	return;
}

1;
