# Copyright (c) 2002 Gavin Brown. All rights reserved. This program is
# free software; you can redistribute it and/or modify it under the same
# terms as Perl itself. 
#
# $Id: WeekDays.pm,v 1.1 2002/11/27 14:57:56 jodrell Exp $

package Gnome::Crontab::WeekDays;

sub new {
	my ($class, $gc, $value) = @_;
	my $gcm = {};
	bless($gcm, $class);

	$gcm->{gc} = $gc;

	my @ary = split(/,/, $value);

	if ($value =~ /\*/) {
		$gcm->{wildcard} = 1;
	} else {
		my @values;
		if ($value =~ /-/) {
			my ($first, $last) = split(/-/, $value, 2);
			my $found_first = 0;
			foreach (qw(mon tue wed thu fri sat sun)) {
				$found_first = 1 if ($_ eq $first);
				push(@values, $_) if ($found_first == 1);
				last if ($_ eq $last);
			}
		} else {
			@values = split(/,/, $value);
		}

		foreach (@values) {
			my $weekday = $gc->string($_);
			$gcm->{values}{$weekday} = 1;
		}
	}

	$gcm->{widget} = Gtk::Table->new(2, 7);
	$gcm->{widget}->set_row_spacings(8);
	$gcm->{widget}->set_col_spacings(8);
	my $column = 0;
	foreach (qw(mon tue wed thu fri sat sun)) {
		my $weekday = $gc->string($_);
		$gcm->{checkboxes}{$weekday} = Gtk::CheckButton->new($weekday);
		$gcm->{checkboxes}{$weekday}->set_active(1) if ($gcm->{values}{$weekday} == 1 || $gcm->{wildcard} == 1);
		$gcm->{checkboxes}{$weekday}->signal_connect('toggled', sub { $gcm->weekday_toggle($weekday) });
		$gcm->{widget}->attach_defaults($gcm->{checkboxes}{$weekday}, $column, $column+1, 0, 1);
		$column++;
	}

	$gcm->{toggle} = Gtk::CheckButton->new($gc->string('select_all'));
	$gcm->{toggle}->set_active(1) if ($gcm->{wildcard} == 1);

	$gcm->{toggle}->signal_connect('toggled', sub { $gcm->select_all_toggle($gc) });

	$gcm->{widget}->attach_defaults($gcm->{toggle}, 0, 7, 1, 2);

	return $gcm;
}

sub weekday_toggle {
	my ($gcm, $weekday) = @_;

	if (!$gcm->{checkboxes}{$weekday}->get_active()) {
		undef $gcm->{wildcard};
		$gcm->{toggle}->set_active(0);
	}

	return;
}

sub select_all_toggle {
	my ($gcm, $gc) = @_;
	if ($gcm->{toggle}->get_active()) {
		foreach (qw(mon tue wed thu fri sat sun)) {
			my $weekday = $gc->string($_);
			$gcm->{checkboxes}{$weekday}->set_active(1);
		}
		$gcm->{wildcard} = 1;
	}
	return;
}

sub value {
	my $gcm = shift;
	if ($gcm->{wildcard} == 1) {
		return '*';
	} else {
		my @return;
		foreach (qw(mon tue wed thu fri sat sun)) {
			my $weekday = $gcm->{gc}->string($_);
			push(@return, $weekday) if ($gcm->{checkboxes}{$weekday}->get_active());
		}
		return lc(join(',', @return));
	}
}

1;